; NVRAM.A51 -- nonvolatile RAM interface

;--- constants

%INCLUDE      "defines.inc"
%INCLUDE      "macros.inc"

;--- external addresses

              EXTERN    _SimMode(BIT)
              EXTERN    load_WR0(CODE)
              EXTERN    load_WR1(CODE)

              EXTERN    TmrDelay100us

;--------------
; NV RAM commands
; The '1' start bit is supplied by the transmitting routine

NVR_READ      EQU       $80                 ; address in low 6 bits
NVR_WRITE     EQU       $40                 ; address in low 6 bits
NVR_ERASE     EQU       $C0                 ; address in low 6 bits
NVR_EWEN      EQU       $30                 ; no address
NVR_EWDS      EQU       $00                 ; no address
NVR_ERAL      EQU       $20                 ; no address
NVR_WRAL      EQU       $10                 ; no address (mfg only...)

NVRSIZE       EQU       64                  ; size of NV RAM in entries


;--------------
; Set and clear bits in 8255 data port
; Presume P2:R0 is loaded with the port addres
; Note the naming convention for bits... b&MASK and b&FLIP

SetPBit      %MACRO     b
              MOVX      A,@R0               ; fetch controls
              ORL       A,#b&MASK           ; bit ON
              XRL       A,#b&FLIP           ; flip if needed
              MOVX      @R0,A               ; set controls again
             %ENDM

ClrPBit      %MACRO     b
              MOVX      A,@R0               ; fetch controls
              ANL       A,#NOT b&MASK       ; bit OFF
              XRL       A,#b&FLIP           ; flip if needed
              MOVX      @R0,A               ; set controls again
             %ENDM


;--------------
; Hard constants
; These presume they'll appear in an 8255 Port C, with split I and O
; There is but one address that must handle reading and writing

CSMASK        EQU       $40                 ; NV RAM chip select bit
CSFLIP        EQU       $00                 ;  invert if needed

SKMASK        EQU       $20                 ; NV RAM serial clock bit
SKFLIP        EQU       $00                 ;  invert if needed

DIMASK        EQU       $10                 ; NV RAM data in (into NVRAM)
DIFLIP        EQU       $00                 ;  invert if needed

DOMASK        EQU       $01                 ; NV RAM data out (from NVRAM)
DOFLIP        EQU       $00                 ;  invert if needed
DOSHIFT       EQU       0                   ; number of right shifts bit..0

MAXTIMEOUT    EQU       100                 ; max timeout @ 100 us/count

;--------------
; Soft constants

              SEG       Constants
              ORG       CONSTBASE+NVR_CONST

I8255Port     DW        $E002               ; I/O port address

I8255CtlAddr  DW        $E003               ; control port address

I8255Mode     DB        10010001B           ; A in, B out, C7:4 out, C3:0 in
              DB        0                   ; padding



;-------------
; Initialize non-volatile RAM hardware

              SEG       UtilCode

_NvrInit      PROC
              PUBLIC    _NvrInit

              GetCData  I8255Mode           ; 8255 control mode
              GetCWord  I8255CtlAddr        ; control reg address

              MOVX      @DPTR,A             ; changing mode resets outputs

              GetCWord  I8255Port           ; 8255 data port address
              MOV       P2,A
              MOV       R0,A

              MOVX      A,@R0               ; fetch current bits

              ANL       A,#NOT CSMASK       ; shut off our controls
              XRL       A,#CSFLIP
              MOVX      @R0,A               ;  (ensure CS is off first)

              ANL       A,#NOT SKMASK
              XRL       A,#SKFLIP
              ANL       A,#NOT DIMASK
              XRL       A,#DIFLIP

              MOVX      @R0,A               ; the rest are just for nice

              MOV       A,#NVR_EWDS         ; disable erase & write commands
              CALL      NvrSendCmd

              RET

_NvrInit      ENDPROC

;-------------
; Read NV RAM data into External RAM
; void NvrGetData(int NvrAddress,int EntryCount,WORD *pNvrData);

              SEG       UtilCode

_NvrGetData   PROC
              PUBLIC    _NvrGetData

              MOV       DPTR,#0             ; fetch count from stack
              CALL      load_WR0            ;   to R0:R1 (only use R1)
              MOV       A,R1                ; skip if count is zero
              JZ        L?done

              MOV       DPTR,#2             ; fetch target pointer
              CALL      load_WR1            ;   to R2:R3

              GetCWord  I8255Port           ; set port address in P2:R0
              MOV       P2,DPH
              MOV       R0,DPL

L?next
              MOV       A,R5                ; read data entry
              ORL       A,#NVR_READ
              CALL      NvrSendCmd          ; activate chip

              MOV       DPH,R2
              MOV       DPL,R3
              CALL      NvrReadData         ; first byte
              MOVX      @DPTR,A
              INC       DPTR
              CALL      NvrReadData         ; second byte
              MOVX      @DPTR,A
              INC       DPTR
              MOV       R2,DPH
              MOV       R3,DPL

              ClrPBit   CS                  ; end of entry

              INC       R5                  ; tick entry number up
              DJNZ      R1,L?next           ; tick count down

L?done
              RET

_NvrGetData   ENDPROC



;-------------
; Write NV RAM page from External RAM image
; void NvrPutData(int NvrAddress,int EntryCount,WORD *pNvrData);

              SEG       UtilCode

_NvrPutData   PROC
              PUBLIC    _NvrPutData

              MOV       DPTR,#0             ; fetch count from stack
              CALL      load_WR0            ;   to R0:R1 (only use R1)
              MOV       A,R1                ; skip if count is zero
              JZ        L?done

              MOV       DPTR,#2             ; fetch source pointer
              CALL      load_WR1            ;   to R2:R3

              GetCWord  I8255Port           ; set port address in P2:R0
              MOV       P2,DPH
              MOV       R0,DPL

              MOV       A,#NVR_EWEN         ; enable read/writes
              CALL      NvrSendCmd
              ClrPBit   CS

L?next
              MOV       A,R5                ; erase entry
              ORL       A,#NVR_ERASE
              CALL      NvrSendCmd
              CALL      NvrWait

              MOV       A,R5                ; send write command
              ORL       A,#NVR_WRITE
              CALL      NvrSendCmd

              MOV       DPH,R2
              MOV       DPL,R3
              MOVX      A,@DPTR             ; fetch first data byte
              CALL      NvrSendData
              INC       DPTR

              MOVX      A,@DPTR             ; fetch second data byte
              CALL      NvrSendData
              INC       DPTR
              MOV       R2,DPH
              MOV       R3,DPL

              CALL      NvrWait             ; wait for RAM to catch up

              INC       R5                  ; tick entry number up
              DJNZ      R1,L?next           ; tick count down

              MOV       A,#NVR_EWDS         ; disable read/writes
              CALL      NvrSendCmd
              ClrPBit   CS

L?done
              RET

_NvrPutData   ENDPROC


;-------------
; Send command and address from A
; Presumes P2:R0 points to 8255 port
; Leaves chip enabled, ready for data I/O
;  ... so you have to finish the command sequence
; Drops NVR DI line after lowering the strobe
; "Data" entry point sends ACC as raw bit stream

NvrSendCmd    PROC
              PUBLIC    NvrSendCmd

              PUSH      ACC                 ; save command byte

              SetPBit   CS                  ; enable the chip

              CLR       C                   ; pre-start bit
              CALL      NvrBitOut

              SETB      C                   ; start bit
              CALL      NvrBitOut

              POP       ACC                 ; recover the byte

;--- entry point for raw byte output from A

NvrSendData   EQU       $
              MOV       B,#8                ; eight cmd + address bits

L?next
              RLC       A                   ; get next bit into C
              CALL      NvrBitOut
              DJNZ      B,L?next

              ClrPBit   DI                  ; drop data to RAM for neatness

              RET

NvrSendCmd    ENDPROC


;-------------
; Read one byte of NV RAM data
; Presumes chip is waiting to send data from port
; Returns byte in A
; Crunches B
; REMEMBER TO READ BOTH BYTES!

NvrReadData   PROC
              PUBLIC    NvrReadData

              MOV       B,#8                ; eight data bits
              CLR       A                   ; start with nothing

L?next
              CALL      NvrBitIn            ; fetch bit into C
              RLC       A                   ; stick bit into A.0
              DJNZ      B,L?next

              RET

NvrReadData   ENDPROC


;-------------
; Wait for end of NV RAM cycle
; Presumes chip is enabled and ready to go
; Returns if the ready line is stuck in BUSY
;  max delay is 10 ms, typically 7-8 ms
; CS must be high from previous operation...

NvrWait       PROC
              PUBLIC    NvrWait

              PUSH      ACC                 ; save bystanders
              PUSH      B
              MOV       B,#MAXTIMEOUT

              ClrPBit   CS                  ; start NVRAM internal timing
              SetPBit   CS                  ; and activate DO status output

              JB        _SimMode,L?done     ; skip all delays in simulation

L?punt
              MOV       A,#1                ; force a known delay per loop
              CALL      TmrDelay100us
              DEC       B
              MOV       A,B
              JZ        L?done              ; bail out if not ready yet

              MOVX      A,@R0               ; fetch DO as status bit
              ANL       A,#DOMASK           ; isolate it
              XRL       A,#DOFLIP           ; flip if needed
              JZ        L?punt              ; zero implies not ready yet

L?done
              ClrPBit   CS                  ; disable CS to save power
              POP       B
              POP       ACC

              RET


NvrWait       ENDPROC


;-------------
; Write single NV RAM bit from C
; Presumes P2:R0 contains 8255 data port address

NvrBitOut     PROC

              PUSH      ACC                 ; save bystander

              MOVX      A,@R0               ; fetch current bits
              ORL       A,#DIMASK           ; presume bit is on
              JC        L?biton
              ANL       A,#NOT DIMASK       ; it's off
L?biton
              XRL       A,#DIFLIP           ; flip if needed
              MOVX      @R0,A               ; set up the data

              SetPBit   SK                  ; shift clock active
              ClrPBit   SK                  ;  ... and inactive

              POP       ACC                 ; recover bystander
              RET

NvrBitOut     ENDPROC


;-------------
; Read single NV RAM bit into C
; Presumes P2:R0 contains 8255 data port address

NvrBitIn      PROC

              PUSH      ACC                 ; save bystander

              SetPBit   SK                  ; shift clock active

              MOVX      A,@R0               ; fetch the bit
              ANL       A,#DOMASK           ; isolate it
              XRL       A,#DOFLIP           ; flip if needed

             %REPT      DOSHIFT
              RR        A                   ; align to bit 0
             %ENDREPT

              MOV       C,ACC.0             ; get bit into carry flag

              ClrPBit   SK                  ; shift clock inactive

              POP       ACC                 ; recover bystander
              RET

NvrBitIn      ENDPROC

              END
