/* LCD.C                                                                     */

/*---------------------------------------------------------------------------*/
/* Macro to simplify command output without actually doing any work...       */

#define LCDSendCmd(x) LCDSendByte(x,LCDWAIT|LCDCMD)


/*---------------------------------------------------------------------------*/
/* A simple single-stepper                                                   */

WaitKey(pMsg)
char *pMsg;
{

   Nputstr(pMsg);
   Nputstr("... ?:");
   while (!chkch());
   Nputstr("\n");
   getch();
}


/*---------------------------------------------------------------------------*/
/* Wait for Busy flag to clear                                               */
/* Need to read both nibbles to get the flag                                 */
/* Returns only the address part of the byte, as busy is always off...       */
/* This bails out after an excessive amount of waiting and sets a flag       */
/*  to disable further delays                                                */

int LCDWaitBusy() {
unsigned Data;

   Data = 0;
   TESTBIT(TestBits,FlagLCDFailed);    /* LCD OK?                            */
   if (!TestBits) {                    /* yes, so do the wait                */
      TickCtr = 0;                     /* reset the timeout counter          */
      do {
         CPLBIT(MAINFLAG);             /* indicate we are spinning           */
         Data = LCDReadByte(LCDCMD);   /* fetch busy/address status          */
         CPLBIT(MAINFLAG);             /* but always restore to known state  */
         if ((Data & 0x80) && (ReportMode & REPORT_LCD)) {
            Nputch('w');
         }
         if (TickCtr > 10) {           /* if timed out, bail out             */
            SETBIT(FlagLCDFailed);
            Data = 0;                  /* force non-busy return code         */
            break;
         }
      } while (Data & 0x80);           /* while still busy                   */
   }

   return Data;

}



/*---------------------------------------------------------------------------*/
/* Read data byte from controller                                            */
/* Remmber that the LCD cursor position advances when you read data          */

int LCDReadByte(Mode)
unsigned Mode;
{
BYTE Data;

   SETBIT(FlagPortBusy);               /* disable timer scans                */

   LCD_PORT |= 0x0F;                   /* set up port for input              */

   if (LCDDATA == Mode) {              /* reading data or address/busy?      */
      SETBIT(LCD_RS);
   }
   else {
      CLRBIT(LCD_RS);
   }

   SETBIT(LCD_RW);                     /* set up LCD for output              */

   SETBIT(LCD_ENABLE);                 /* turn on the strobe                 */
   Data = LCD_PORT & ~INPUTMASK;       /* fetch and isolate high nibble      */
   Data <<= 4;                         /* align it in the return byte        */
   CLRBIT(LCD_ENABLE);                 /* indicate we got it                 */

   SETBIT(LCD_ENABLE);                 /* turn on the strobe                 */
   Data |= LCD_PORT & ~INPUTMASK;      /* fetch and isolate low nibble       */
   CLRBIT(LCD_ENABLE);                 /* indicate we got it                 */

   CLRBIT(FlagPortBusy);               /* OK to tickle the port again        */

   return Data;

}


/*---------------------------------------------------------------------------*/
/* Send command or data byte to the LCD controller                           */
/* If this is data, the LCD cursor position will advance                     */
/* Presumes 4-bit interface mode...                                          */

LCDSendByte(Data,Mode)
unsigned Data;
unsigned Mode;
{
BYTE OutByte;

   if (!(Mode & LCDFORCE)) {           /* want to force without wait?        */
      LCDWaitBusy();
   }

   SETBIT(FlagPortBusy);               /* disable timer scans                */

   if (Mode & LCDDATA) {               /* set up for command or data byte    */
      SETBIT(LCD_RS);
   }
   else {
      CLRBIT(LCD_RS);
   }

   CLRBIT(LCD_RW);                     /* set up to write the data           */

   OutByte = (LCD_PORT & INPUTMASK) | ((Data >> 4) & ~INPUTMASK);
   LCD_PORT = OutByte;                 /* high nibble goes first             */
   SETBIT(LCD_ENABLE);                 /* enable to write                    */
   CLRBIT(LCD_ENABLE);

   OutByte = (LCD_PORT & INPUTMASK) | (Data & ~INPUTMASK);
   LCD_PORT = OutByte;                 /* low nibble goes last               */
   SETBIT(LCD_ENABLE);                 /* enable to write                    */
   CLRBIT(LCD_ENABLE);

   SETBIT(LCD_RW);                     /* disable write line                 */

   CLRBIT(FlagPortBusy);               /* OK to tickle the port again        */

}




/*---------------------------------------------------------------------------*/
/* Initialize LCD hardware and display variables                             */

LCDInitialize() {

   if (ReportMode & REPORT_LCD) {
      Nputstr("--- Initializing LCD...");
   }

   LCDFlags = 0;                       /* flush bit flags to zero            */

   CurrentRow = CurrentCol = 0;

   LCDControls = ControlDefs;

   LCD_PORT |= ~INPUTMASK;             /* float all data lines               */
   CLRBIT(LCD_ENABLE);                 /* turn off enable line               */
   CLRBIT(LCD_RS);                     /* set command mode                   */
   CLRBIT(LCD_RW);                     /* set write mode                     */

/*--- first few commands cannot check BUSY, as state is not known            */

   SETBIT(FlagPortBusy);               /* disable timer scans                */

   TmrDelayTicks(15/MSPERTICK);        /* delay for >15 ms                   */
   LCD_PORT = 0x03 | (LCD_PORT & INPUTMASK);
   SETBIT(LCD_ENABLE);
   CLRBIT(LCD_ENABLE);

   TmrDelayTicks(5/MSPERTICK);         /* delay for >4.1 ms                  */
   LCD_PORT = 0x03 | (LCD_PORT & INPUTMASK);
   SETBIT(LCD_ENABLE);
   CLRBIT(LCD_ENABLE);

   TmrDelay100US(2);                   /* delay for >100 us (easy!)          */
   LCD_PORT = 0x03 | (LCD_PORT & INPUTMASK);
   SETBIT(LCD_ENABLE);
   CLRBIT(LCD_ENABLE);

   TmrDelayTicks(1);                   /* unspecified delay                  */

   LCD_PORT = 0x02 | (LCD_PORT & INPUTMASK);       /* set 4-bit interface    */
   SETBIT(LCD_ENABLE);
   CLRBIT(LCD_ENABLE);

   CLRBIT(FlagPortBusy);               /* enable timer fiddling              */

/*--- OK to check BUSY for the remaining setup commands                      */

   LCDSendCmd(0x28);                               /* 4 bits, 2 lines, 5x7   */
   LCDSendCmd(0x08);                               /* disp/cur/blink off     */
   LCDSendCmd(0x01);                               /* clear and home         */
   LCDSendCmd(0x06);                               /* increment, no shift    */

   CursorState = CursorDef | 0x0C;                 /* set up cursor          */
   LCDSendCmd(CursorState);                        /* ... with display ON    */

   if (ReportMode & REPORT_LCD) {
      Nputstr(" done\n");
   }

}


/*---------------------------------------------------------------------------*/
/* Set LCD cursor to row/column location                                     */
/* Rows and columns start from 0 in the upper-left corner                    */

LCDSetCursor(Row,Col)
int Row;
int Col;
{
BYTE CursorCmd;

   Row = max(Row,0);                   /* prevent wrap off top               */
   Col = max(Col,0);

   CurrentRow = Row = min(Row,VisibleRows-1);      /* and off bottom         */
   CurrentCol = Col = min(Col,VisibleCols-1);

   CursorCmd = RowStarts[Row] + Col | 0x80;        /* 0x80 sets cursor       */

   if (ReportMode & REPORT_LCD) {
      sprintf(CmdRespBuff,"--- LCD to [%u %u], cmd=%02x\n",
              Row,Col,CursorCmd);
      Nputstr(CmdRespBuff);
   }

   LCDSendCmd(CursorCmd);

}


/*---------------------------------------------------------------------------*/
/* Set cursor to 0,0 but do not clear the display                            */

LCDHome() {

   CurrentRow = CurrentCol = 0;

   LCDSendCmd(0x02);

}


/*---------------------------------------------------------------------------*/
/* Clear display and home the cursor                                         */

LCDClear() {

   CurrentRow = CurrentCol = 0;

   LCDSendCmd(0x01);

}


/*---------------------------------------------------------------------------*/
/* Scroll display up N lines                                                 */

LCDScrollUp(NumLines)
unsigned NumLines;
{
BYTE SaveRow;
BYTE SaveCol;
BYTE ThisRow;
BYTE PrevRow;
BYTE ThisCol;

   if (ReportMode & REPORT_LCD) {
      sprintf(CmdRespBuff,"LCD scroll up %u...  ",NumLines);
      Nputstr(CmdRespBuff);
   }

   if (!NumLines) {                    /* this makes no sense...             */
      return;
   }

   if (NumLines >= VisibleRows) {      /* this is easy                       */
      LCDClear();
      return;
   }

   if (LCDControls & LCD_BLANKING) {   /* blank during scrolling?            */
      LCDSelDisplay(LCDDISPOFF);
   }

   LCDWaitBusy();
   SaveRow = CurrentRow;
   SaveCol = CurrentCol;


   for (ThisRow = NumLines, PrevRow = 0;           /* from and to targets    */
        ThisRow < VisibleRows;
        ++ThisRow, ++PrevRow) {
      LCDSetCursor(ThisRow,0);                     /* fetch whole row        */
      for (ThisCol = 0; ThisCol < VisibleCols; ++ThisCol) {
         RowBuffer[ThisCol] = LCDReadByte(LCDDATA);
      }
      LCDSetCursor(PrevRow,0);                     /* move to target         */
      for (ThisCol = 0; ThisCol < VisibleCols; ++ThisCol) {
         LCDSendByte(RowBuffer[ThisCol],LCDWAIT|LCDDATA);
      }
   }

   LCDSetCursor(VisibleRows-1,0);                  /* set last row to blank  */
   for (ThisCol = 0; ThisCol < VisibleCols; ++ThisCol) {
      LCDSendByte(' ',LCDWAIT|LCDDATA);
   }

   LCDSetCursor(SaveRow,SaveCol);                  /* restore cursor loc     */
   LCDSelDisplay(LCDDISPON);                       /* always enable display  */

   if (ReportMode & REPORT_LCD) {
      Nputstr("... scroll done\n");
   }


}


/*---------------------------------------------------------------------------*/
/* Enable/disable LCD cursor                                                 */

LCDCtlCursor(State)
unsigned State;
{
   switch (State) {
   case LCDCURSOFF :
      CursorState = 0;
      break;
   case LCDCURSON :
      CursorState = CursorDef;
      break;
   }

   LCDSendCmd(CursorState);


}



/*---------------------------------------------------------------------------*/
/* Enable/disable LCD panel                                                  */

LCDSelDisplay(State)
unsigned State;
{
unsigned DispCmd;

   DispCmd = CursorState | 0x08;                   /* assume display OFF     */

   DispCmd |= (LCDDISPON == State) ? 0x04 : 0;     /* set ON or OFF          */

   LCDSendCmd(DispCmd);


}



/*---------------------------------------------------------------------------*/
/* Send printable chars to the LCD                                           */
/* Handles standard control chars                                            */

LCDSendChar(Char)
char Char;
{
BYTE ThisCol;
BYTE ThisChar;

   if (ReportMode & REPORT_LCD) {
      sprintf(CmdRespBuff,"LCD %02x [%c]\n",
              Char,iscntrl(Char) ? '-' : Char);
      Nputstr(CmdRespBuff);
   }

   switch (Char) {
   case LF :
      if (CurrentRow == (VisibleRows - 1)) {       /* on last row?           */
         if (LCDControls & LCD_VSCROLL) {          /* yes, want scrolling?   */
            LCDScrollUp(1);                        /* yes, do it & stay put  */
         }
         else {                                    /* no scrolling           */
            CurrentRow = 0;                        /*  so return to top      */
         }
      }
      else {                                       /* not on last row        */
         ++CurrentRow;                             /* so just step down one  */
      }
      if (!(LCDControls & LCD_CRLF)) {             /* want CR after newline? */
         LCDSetCursor(CurrentRow,CurrentCol);      /*  ... no, so done       */
         break;
      }
      /* else fall through */
   case CR :
      LCDSetCursor(CurrentRow,0);                  /* to leftmost column     */
      break;
   case FORMFEED :
      LCDClear();
      break;
   case BS :
      if (CurrentCol) {                            /* if not at edge         */
         --CurrentCol;                             /*   back up one col      */
      }
      if (!(LCDControls & LCD_DESTBS)) {           /* want char destruction? */
         break;
      }
      /* else fall through */
   case DEL :
      LCDWaitBusy();
      if (LCDControls & LCD_BLANKING) {            /* blank during scroll?   */
         LCDSelDisplay(LCDDISPOFF);
      }

      if (ReportMode & REPORT_LCD) {
         Nputstr("  Readback chars are [");
      }
      for (ThisCol = CurrentCol; ThisCol <= VisibleCols; ++ThisCol) {
         ThisChar = RowBuffer[ThisCol] = LCDReadByte(LCDDATA);
         if (ReportMode & REPORT_LCD) {
            Nputch(isprint(ThisChar) ? ThisChar : '-');
         }
      }
      if (ReportMode & REPORT_LCD) {
         Nputstr("]\n");
      }

      LCDSetCursor(CurrentRow,CurrentCol);         /* restore cursor loc     */

      if (ReportMode & REPORT_LCD) {
         Nputstr("  Reloaded chars are [");
      }
      for (ThisCol = CurrentCol+1; ThisCol < VisibleCols; ++ThisCol) {
         ThisChar = RowBuffer[ThisCol];            /* slide over 1           */
         LCDSendByte(ThisChar,LCDWAIT|LCDDATA);
         if (ReportMode & REPORT_LCD) {
            Nputch(isprint(ThisChar) ? ThisChar : '-');
         }
      }
      LCDSendByte(' ',LCDWAIT|LCDDATA);
      if (ReportMode & REPORT_LCD) {
         Nputstr(" ]\n");
      }
      LCDSetCursor(CurrentRow,CurrentCol);         /* restore cursor loc     */
      LCDSelDisplay(LCDDISPON);                    /* always enable display  */
      break;
   default :
      LCDSendByte(Char,LCDWAIT|LCDDATA);           /* send the char          */
      ++CurrentCol;                                /* tick column counter    */
      if (CurrentCol >= VisibleCols) {             /* past end of line?      */
         LCDSetCursor(CurrentRow,0);               /* reset to left edge     */
         if (!(LCDControls & LCD_HWRAP)) {         /* go to next line?       */
            LCDSendChar(LF);                       /* yes, so force it       */
         }
      }
   }

}




/*---------------------------------------------------------------------------*/
/* Send C string to the LCD                                                  */

LCDSendString(pString)
char *pString;
{

   while (0 != *pString) {
      LCDSendChar(*pString);
      ++pString;
   }

}



