/* TIMER.C                                                                   */

/*---------------------------------------------------------------------------*/
/* Set up timers and suchlike                                                */

TmrInitialize() {

/*--- interrupt vectors                                                      */

 asm {
IRHERE  EQU *
   ORG   $000B                   /* Timer 0 interrupt     (IR pulses)        */
   LJMP  Tmr0Tick
   ORG   IRHERE
 }

/*--- global variables                                                       */

   TmrBits = 0;                        /* flush all flags at once            */

   T0Reload = T0RldNorm;               /* set up normal interrupt period     */

   T0Ticks = 0;                        /* start from scratch                 */
   TickCtr = 0;                        /* not really needed                  */
   BeepCtr = 0;                        /* beep stays off for now...          */

   KeysHit = KeysLast = 0;             /* set up key test memories           */

/*--- Set up Timer 0 for long-period timings                                 */

   TCON &= 0xCF;                 /* clear run & interrupt flags              */

   TMOD &= 0xF0;                 /* set mode 1 (16-bit reload)               */
   TMOD |= 0x01;

   TH0 = TL0 = 0;                /* first delay is a long one                */

   SETBIT(PT0);                  /* high priority                            */
   SETBIT(ET0);
   SETBIT(TR0);                  /* start it running                         */

}


/*---------------------------------------------------------------------------*/
/* Timer 0 interrupt handler                                                 */
/* This is in assembler just to save some overhead                           */

Tmr0Tick() {

;
 asm {
*  SETB  TIMERFLAG
   PUSH  A
   PUSH  PSW

*--- reload the timer

   CLR   TR0
   MOV   TL0,T0Reload
   MOV   TH0,T0Reload+1
   SETB  TR0

*--- save bystanders for the rest of the interrupt

   PUSH  B
   PUSH  DPH
   PUSH  DPL
   PUSH  0                             ; R0 in disguise

*--- check for any key presses and remember which ones are active
*    need to preserve outputs to LEDs while checking inputs from switches

   JB    FlagPortBusy,swdone           ; skip if port is in use

   MOV   B,#$FF                        ; set up output latch memory

   JBC   SWITCH_PORT.7,sw?7            ; branch if latch was set, clear latch
   CLR   B.7                           ;  latch was off, record in B
sw?7  EQU   *
   JBC   SWITCH_PORT.6,sw?6
   CLR   B.6
sw?6  EQU   *
   JBC   SWITCH_PORT.5,sw?5
   CLR   B.5
sw?5  EQU   *
   JBC   SWITCH_PORT.4,sw?4
   CLR   B.4
sw?4  EQU   *
   JBC   SWITCH_PORT.3,sw?3
   CLR   B.3
sw?3  EQU   *
   JBC   SWITCH_PORT.2,sw?2
   CLR   B.2
sw?2  EQU   *
   JBC   SWITCH_PORT.1,sw?1
   CLR   B.1
sw?1  EQU   *
   JBC   SWITCH_PORT.0,sw?0
   CLR   B.0
sw?0  EQU   *

   MOV   SWITCH_PORT,B                 ; restore latches
   ORL   SWITCH_PORT,#INPUTMASKASM     ; set input bits high for reading
   MOV   A,SWITCH_PORT                 ; fetch inputs

   MOV   SWITCH_PORT,B                 ; restore latches again

   CPL   A                             ; convert to positive logic (1 = ON)
   ANL   A,#INPUTMASKASM               ; isolate input bits
   MOV   B,A                           ; save for later update
   XRL   A,KeysLast                    ; compare with previous sample
   MOV   KeysLast,B                    ; and update previous sample
   JZ    swdone                        ; if zero, nothing changed

   ANL   A,B                           ; 1 bits = new ON buttons
   ORL   KeysHit,A                     ; record the fact...

swdone   EQU   *

*--- tick beep timer and activate the output if nonzero

   MOV   A,BeepCtr
   JZ    ?beep2                        ; zero means beep not in progress

   CLR   OUT_BEEP                      ; turn output on when nonzero
   DJNZ  A,?beep2                      ; count and test for finished

?beep1   EQU   *
   SETB  OUT_BEEP                      ; turn off when we hit zero

?beep2   EQU   *
   MOV   BeepCtr,A

*--- update counter and check for a full second

   INC   TickCtr                       ; increment general timer

   INC   T0Ticks                       ; tick the wrap counter

   MOV   A,T0Ticks                     ; second yet?
   CJB   A,#TICKSPERSEC,T0End

   SETB  FlagSecond                    ; yes, record this fact
   ADD   A,#-TICKSPERSEC               ; and knock off a full second
   MOV   T0Ticks,A

*--- restore regs and return...

T0End    EQU   *
   POP   0                       ; restore R0
   POP   DPL
   POP   DPH
   POP   B
   POP   PSW
   POP   A

*  CLR   TIMERFLAG
   RETI                          ; pre-empt normal C level return

 }

}


/*---------------------------------------------------------------------------*/
/* Delay for a fixed number of milliseconds                                  */
/* Zero NumTicks will wait next tick                                         */

TmrDelayTicks(NumTicks)
unsigned NumTicks;
{
BYTE TickSample;

   do {
      TickSample = T0Ticks;            /* catch current timer tick counter   */
      while (TickSample == T0Ticks);   /* wait for timer tick                */
   } while (NumTicks--);               /* count each tick                    */



}


/*---------------------------------------------------------------------------*/
/* Delay for a fixed number of 100 microsecond ticks                         */
/* Zero NumUS is special-cased to bail out                                   */

TmrDelay100US(Num100US)
unsigned Num100US;
{

   if (!Num100US) {                    /* bail out for zero input            */
      return;
   }

   do {
 asm {
      MOV   B,#46
Tmr?1 EQU   *
      DJNZ  B,Tmr?1
 }
   } while (--Num100US);



}
