$mod31
$pagewidth(125)
$pagelength(84)
$nodebug
;*******************************************************************************
;*                                                                             *
;*        INFRARED REMOTE CONTROL  RECEIVER, DECODER AND TRANSMITTER           *
;*                                                                             *
;* Using a standard surplus IR receiver module, capture and store the codes    *
;* from most of the IR remotes and decode for press and repeat sequences.      *
;* the standard modules convert pulsed 38KHz IR into a series of highs and     *
;* lows, filtering out the 38KHz carrier.  Each manufacturer has different     *
;* methods of encoding the commands into the pulsed IR - PPM, PWM, NRZ, etc.,  *
;* so the decoding routines must be generic enough to handle all the formats   *
;* and lengths.  Simply timing the periods of carrier/no carrier and storing   *
;* the results is the method I used here, and will work on nearly all remotes  *
;* in use except for early ones that didn't use the 38KHz carrier.  Instead,   *
;* they frequently used pulsed series of different frequencies or tones, then  *
;* decoded 'em with a DTMF decoder chip.  The only way to catch one of these   *
;* lousy formats is with an IR receiver that doesn't filter out the carrier,   *
;* and they're scarce as hen's teeth. If I can find one, it'll be added later. *
;*                                                                             *
;* Most remotes send the same code over and over as a repeat function, both to *
;* allow for quick scanning (FF on VCRs) and to repeat the channel up/down and *
;* volume buttons.  A few send one code for the pressed button, then a short   *
;* different pattern to repeat it.  A few only send the code once, most        *
;* continually repeat it until you let the button up.  There's always a short  *
;* delay between codes, either the re-sent code or the repeat code, so you can *
;* use the delays to distinguish between 'em (it's be the longest pause in     *
;* transmission).  Most receivers recognize the correct code the first time,   *
;* but some like the Sony units need the same code 3 times to respond.         *
;* The different pulse widths I've observed varies between 50uS and 100mS,     *
;* quite a range to effectively decode with a timer.  The output pin on the    *
;* modules I have here goes low when receiving a 38KHz carrier, and goes high  *
;* or open-collector when no carrier is present.  This code will work with the *
;* other variety (carrier-high) as long as you don't press the button until    *
;* the routine starts.  The algorithm is designed for around an 11MHz crystal, *
;* and shouldn't need any changes unless you're using a crystal below 3MHz or  *
;* in excess of 16MHz.  You'd need to modify the timer to use an external clk  *
;* if you're outside that range, anyway.                                       *
;*                                                                             *
;* This code is written to work on any 8031 derivative, so all you need to do  *
;* is re-ORG it to wherever you need it or LINK it into existing code and it's *
;* ready to go with no other modifications.  The buffers are insensitive to    *
;* location, and can be moved to wherever you have room.  The use of Timer 0   *
;* can be easily changed to Timer 1 if you don't have an ISR serial port, and  *
;* could use Timer 2 in the 8052 derivatives with some configuration changes.  *
;*                                                                             *
;* For the transmission end, I use a switched oscillator to generate the PCM   *
;* signal to be sent out.  It's easier than trying to time the 26uS period of  *
;* each carrier cycle in software, and allows for a wider range of crystal     *
;* speeds since the carrier is crystal independent.  To send out a particular  *
;* code, you just need to find the length of any sequence and whether there's  *
;* a repeat code, and use the timer to send the pulsed code out the IR LED.    *
;* To make it insensitive to VCC variations, I used a UJT for the oscillator.  *
;*                                                                             *
;* A word of warning - you MUST set the first byte at IRTABL to zero the first *
;* time you use these routines, or it'll take whatever random number is there  *
;* and use the random junk above it as the locations to store the final data   *
;* to.  It's a good idea to set it in the power-on reset routine if the chip   *
;* is constantly powered.  Next thing to add is an editor so you can modify    *
;* weird responses from uncultured remotes to match your expectations.         *
;* If I get REALLY bored later, an IR decoder (to button #) will be added.     *
;* A note on program flow - both the transmit and learn routines start the     *
;* timer while they're working on the next/last bit processing to cut the time *
;* wasted with the timer off.  Watch how you code around those routines.       *
;* They were carefully crafted to take advantage of the minimum pulse width    *
;* (52uS) that could possibly come out of the receiver module, and the way I   *
;* coded it shouldn't miss anything or artificially stretch the time too much. *
;* Since the carrier has a 26uS period, any timing errors less than that will  *
;* be ignored for in the equipment receiving the pulses (it'd have to).  Also, *
;* note that my carrier is ASYNCHRONOUS to the transmitted code, whereas the   *
;* real remotes are usually synchronous since they divide the 455-480KHz       *
;* ceramic resonator down internally in the remote controller chip.            *
;*                                                                             *
;* I haven't done any testing to see what the allowable error in pulse timing  *
;* is for my equipment, but tried to minimize any gross errors up front.       *
;* Must not be too bad - it's more accurate than my Yamaha MRX master remote.  *
;* Their unit adds spurious pulses occasionally, and has other strange timing  *
;* inconsistencies, so anything I've screwed up here shouldn't matter much.    * 
;*                                                                             *
;*******************************************************************************
$eject
;
;       hardware and program storage definitions
;
XSEG
IR_BUF	EQU	8000h		; 512 byte XDATA buffer for the received code
IRTABL	EQU	IR_BUF+200h	; 16 byte table pointers to location of codes
IRCODE	EQU	IR_BUF+400h	; and lots of room to store the individual codes
				;  may need as much as 8K depending on remotes
DSEG	at	40h
max_delay:	ds	2	; storage to find the longest repeat delay time

BSEG
IR_REPEAT	EQU	$	; send the repeat code? high, yes  low, no


IR_IN	EQU	P3.2		; Port 3.2 has the filtered IR receiver module
LED_EN	EQU	P3.3		; Driving Port 3.3 high turns the IR LEDs on
				; (the carrier is automatically added in hardware)

	CSEG
	ORG	2000h		; not hard-coded (you can change the location)

$eject
;*******************************************************************************
;*                                                                             *
;*  This routine is entered with a single byte in the accumulator to tell it   *
;*  which entry in the stored IR table to transmit.  The only returned codes   *
;*  to the calling program are a 00h if the table has no entries, or an FFh if *
;*  the requested entry is beyond the end.  There's a provision for handling   *
;*  the transmission of the repeat code if needed.  Just have some way the     *
;*  calling program can set and clear the IR_REPEAT flag via an ISR from the   *
;*  keyboard or serial port, and it'll either send out the repeat code or it   *
;*  will repeat the button code if no repeat code exists.  Make sure you have  *
;*  some sort of interrupt handler in place BEFORE you set the IR_REPEAT flag, *
;*  or the hardware will lock up and repeat the button unto eternity.  Even    *
;*  though it shouldn't be needed, the flag is cleared on exit.                *
;*  Also, if you have multiple interrupt sources with >30uS ISRs, you may want *
;*  to re-code the routines and only enable EA when you're transmitting the    *
;*  inter-code repeat delay.  It won't hurt if it's stretched a little, but    *
;*  the bit times shouldn't be changed too much or your equipment may not      *
;*  properly recognize them.  See the area right after repeat_check.           *
;*  For normal switched 38KHz PCM, you wiggle the port pin used for the LED_EN *
;*  either high for carrier or low for no carrier.  
;*  The UJT relaxation oscillator is relatively insensitive to variations in   *
;*  temperature or VCC, and is a better choice than many since the frequency   *
;*  must be 39KHz +/- 500Hz for some of the stock receiver modules, and using  *
;*  a simple R-C oscillator won't cut it.  Nice thing - UJTs are cheap, too.   *
;*                                                                             *
;*******************************************************************************

IR_TRANSMIT:
;	PUSH 	TCON		; save the previous setup of Timer 0
;	PUSH	TMOD
;	PUSH	IE
;	CLR	ET0		; make sure we don't interrupt when timer ovfs
;	PUSH	TH0
;	PUSH	TL0
;	CLR	EA		; or at all
	CLR	LED_EN		; first, turn off the LEDs
	ANL	TCON,#0CCh	; shut down whatever was done with Timer 0
	ANL	TMOD,#0F0h
	ORL	TMOD,#1		; set Timer 0 as 16-bit timer, ungated, 1 clk/cy
	MOV	R7,A		; save for the return value
	MOV	B,A
	MOV	DPTR,#IRTABL	; and find out if it's a valid button
	MOVX	A,@DPTR		; find out how many have been stored
	JNZ	nxtchk		; if there aren't any entries, quit
	AJMP	invalid_button
nxtchk:	CJNE	A,B,chkbyt	; used 'cos Intel didn't do a < > compare jump
				; and instead wasted codes for 8048 compatibility
				; they should have trashed the AJMPs & ACALLs
chkbyt:	JC	invalid_button	; if it's higher than current top, quit too
	DEC	DPL		; index backwards once to get the loop set up
	MOV	A,#0FFh
	CJNE	A,DPL,find_button	; since they didn't include a DEC DPTR
	DEC	DPH
find_button:
	MOV	A,R7		; get the requested button back
find_button_loop:
	INC	DPTR		; and move the DPTR up to the right entry
	INC	DPTR
	DJNZ	ACC,find_button_loop

	MOVX	A,@DPTR		; get the hi byte of the table entry
	MOV	R5,A
	INC	DPTR
	MOVX	A,@DPTR		; and the low byte
	MOV	DPL,A
	MOV	DPH,R5		; DPTR now points to first byte of entry
	MOVX	A,@DPTR		; have the length of the button code
	MOV	R0,A		; used to keep track for repeat (if none)
	INC	DPTR
	MOVX	A,@DPTR		; and the length of the repeat code (0 if none)
	MOV	R2,A
	INC	DPTR		; move up to first byte of IR code
	MOV	R3,DPH		; and save if used for repeat code (I.E. - user
	MOV	R4,DPL		; wants a repeat function, but no repeat defined)
	CLR	F0		; used to find out if first time through loop
	SETB	TF0		; first time in loop, fall through the JNB

;		
; got all the relevant information now.  Let's start sending the button code out.
;

outer_loop:			; used to reset DPTR and re-initialize counts
	MOV	DPH,R3		; when we're sending repeat codes out
	MOV	DPL,R4
	MOV	A,R0		; keep track of how many bits sent for this code
	MOV	R1,A
xmit_loop:
	MOVX	A,@DPTR		; get the packed timer hi load value
	MOV	B,A
	INC	DPTR
	MOVX	A,@DPTR		; and the packed timer lo load value
	INC	DPTR
	CLR	C
	RLC	A		; unpack low byte 
	XCH	A,B
	RLC	A		; unpack high byte, leaving overflow flag in carry
	JNB	TF0,$		; wait here for timer to finish unless first entry
	CLR	TR0
	CLR	TF0
;
	CPL	LED_EN		; toggle the IRLED state (off/carrier)
;
	JNC	not_ovf		; if the timer didn't overflow, just load values
	MOV	TH0,ACC		; else load value and run once
	MOV	TL0,B		; (done here 'cos might have barely overflowed)
	SETB	TR0
	MOV	A,#0		; set to max time for next period since it
	MOV	B,#0		;  overflowed once during capture
	JNB	TF0,$		; wait here for the timer to time out
	CLR	TR0
	CLR	TF0
not_ovf:
	MOV	TH0,A		; set the timer with the unpacked count
	MOV	TL0,B
	SETB	TR0		; and start it up again

	DJNZ	R1,xmit_loop	; continue for all bits in the code

repeat_check:
	MOV	A,R7		; get the button number back if quitting

	JNB	IR_REPEAT,ir_tx_wait	; if no repeat required, quit
	MOV	A,R2
	JZ	outer_loop	; if no repeat code exists, use button code
	JB	F0,outer_loop	; else use the start of the repeat and loop
	MOV	R0,A		; save repeat code as new length
	MOV	R3,DPH		; and save the start of repeat as new start
	MOV	R4,DPL	
	SETB	F0
	JMP	outer_loop	; until IR_REPEAT drops low from your ISR

IR_tx_wait:
	JNB	TF0,$		; wait for the last bit to finish up
	CLR	TR0
	SJMP	IR_tx_ret	; and quit

invalid_button:			; error in the request - leave result in acc
	JZ	IR_tx_ret	; if there weren't any entries, just quit
	MOV	A,#0FFh		; else tell 'em they're beyond legal button #s
IR_tx_ret:	
;	POP 	TL0		; restore the previous setup of Timer 0
;	POP	TH0
;	POP	IE
;	POP	TMOD
;	POP	TCON
	CLR	LED_EN		; then turn off the LEDs and oscillator
;
	RET		; go on back to the calling program

$eject
;*******************************************************************************
;*                                                                             *
;*  This is a rather simple routine to use one of the timers in 16-bit timer   *
;*  mode to capture an IR bit pattern, decode it as button press and repeat    *
;*  codes (it doesn't check to see if they're different), and finally store    *
;*  the result in a packed pair of tables.  The code is totally insensitive    *
;*  to what flavor of receiver module you have, and even open-drain units will *
;*  work with the port pullup resistor.  The entries saved in the table will   *
;*  always be carrier time, no carrier, carrier, etc. as long as the routine   *
;*  is entered BEFORE the button is pressed.  It's the simplest thing I could  *
;*  do to get the job done, and doesn't take any precautions to see if you've  *
;*  already punched the same button, so it's up to the calling program to      *
;*  insure that we don't fill the tables with junk and duplicate buttons.      *
;*  It ends with a RET as it's meant to be a subroutine from a main program.   *
;*  One recommendation for modifications - I pre-diddle the values that will   *
;*  be loaded into TH0 & TL0 during the transmission, and that makes it kinda  *
;*  hard to read if you're hand editing a stored sequence.  It saves some time *
;*  later (maybe needlessly) yet may cause misunderstanding.  It's much more   *
;*  intuitive if you just use the raw numbers saved initially as they increase *
;*  with longer bit times.  If you add an editor, please clean up the format   *
;*  and re-convert to the final timer reload values when you're all done.      *
;*  It doesn't bother me, but I'm used to thinking sdrawkcab.                  *
;*                                                                             *
;*******************************************************************************

IR_LEARN:

	MOV	SP,#50h		; had to move it up - monitor pgm trashed R6 & R7
				; when I left the stack at 7.  S/B ok in real 552
	PUSH 	TCON		; save the previous setup of Timer 0
	PUSH	TMOD
	PUSH	IE		; and the interrupt status
	PUSH	TH0
	PUSH	TL0
	CLR	ET0		; make sure we don't interrupt when timer ovfs
	ANL	TCON,#0CCh	; shut down whatever was done with Timer 0
	ANL	TMOD,#0F0h
	ORL	TMOD,#1		; set Timer 0 as 16-bit timer, ungated, 1 clk/cy

	CLR	TR0		; disable the timer for the moment if it's running
	MOV	TMOD,#01h	; set timer to Mode 1 (16 bit timer)
	CLR	TF0		; clear any previous overflow

	CLR	A		; clear out the storage area first by wiping
	MOV	DPTR,#IR_BUF	;  all 512 bytes to zeros - don't really need
CLRRAM:	MOVX	@DPTR,A		;  this in a real environment, just did it to
	DJNZ	DPL,CLRRAM	;  make debugging easier.  I wanted to see
	INC	DPH		;  exactly where new data was going to.
CLRAML:	MOVX	@DPTR,A
	DJNZ	DPL,CLRAML
	DEC	DPH

	MOV	R7,#0FFh	; capture the infrared bits - up to 127 bits
				;  or 254 highs and lows ( total 510 bytes)
				;  because we need room for lengths when done
	MOV	TL0,#00		; set to max range for longest bit time possible
	MOV	TH0,#00		; 
	CLR	EA		; turn off interrupts (time critical area next)

	JNB	IR_IN,hi_mod	; decide what kind of module is out there
	SETB	F0		; used to toggle bit-high, bit-low
	JB	IR_IN,$		; wait here for button press - carrier low module
	SJMP	start_IR_learn

hi_mod:	CLR	F0		; this module goes hi for received carrier
	JNB	IR_IN,$

start_IR_learn:
	SETB	TR0		; start the timer to capture the time

$eject
;-------------------------------------------------------------------------------
; The following routine times each high and low time and stores the result in
; XDATA for later processing.  Values are stored in alternating sequence, with
; carrier first, no carrier second, carrier, no carrier, etc.  The routine
; watches both the input pin and the timer overflow flag.  When the input flops,
; it stops the timer and stores the value.  The first time the timer overflows
; (if it does), it sets the carry and continues to look at the input pin.  If 
; the timer overflows a second time on any bit, I assume the transmission has 
; ended and quit to the decode routines.  It'll only overflow twice if the time
; is longer than about 142mS (presuming 11.059MHz xtal), and the max I've seen
; on any remote is less than 100mS so we know the button has been released.
; Likewise, I haven't seen more than 40 bits used for any transmission, so the
; button plus any repeat code should fit within the 128 bit total I've allowed
; for in the buffer size.  When we hit the max count, it quits automatically or
; repeat keys would continue to fill the buffer to infinity.
;-------------------------------------------------------------------------------

get_ir_bit: CLR	C		; carry used to mark timer overflow (long time)
	JNB	F0,ir_hi	; F0 keeps track of what level we're looking for

ir_lo:	JB	IR_IN,got_bit	; as soon as the bit goes high, save the time
	JNB	TF0,ir_lo	; check to see if the timer overflowed
	CLR	TF0		; if it did, clear it
	JB	CY,process_code	;  and see if it overflowed already (done)
	SETB	C		; else mark it for later
	JMP	ir_lo		;  and continue waiting for the bit to rise

ir_hi:	JNB	IR_IN,got_bit	; as soon as the bit goes low, save the time
	JNB	TF0,ir_hi	; check to see if the timer overflowed
	CLR	TF0		; if it did, clear it
	JB	CY,process_code	;  and see if it overflowed already (done)
	SETB	C		; else mark it for later
	JMP	ir_hi		;  and continue waiting for the bit to drop

got_bit: CLR	TR0		; stop the timer briefly to save the values
	
	MOV	R0,TH0		; store 17 bit counter time in the input buffer,
	MOV	R1,TL0		;  shifted so overflow shows up as the msbit
	MOV	TL0,#00		; reload the timer for the next time period
	MOV	TH0,#00		; 
	CLR	TF0		; new bit so ignore previous overflows
	SETB	TR0		; start the timer for the next bit period

	MOV	A,R0		; store the time in the buffer (both bytes)
	RRC	A		; shift in the overflow flag (if any)
	MOVX	@DPTR,A		;  with hi byte first
	MOV	A,R1		;  and low byte second
	RRC	A		; TF1 is new msbit 'cos we need length here, not
	INC	DPTR		;  accuracy.  granularity is still around 2uS.
	MOVX	@DPTR,A	
	INC	DPTR		; move the buffer pointer up for the next bit
	CPL	F0		; toggle the bit hi/low flag
	DJNZ	R7,get_ir_bit	; and continue till all 127 bits or double ovf
				; (overflow twice meant REAL long time - done)

$eject
;-------------------------------------------------------------------------------
; Now process the data and split into codes and repeat sequences (if any).
; If the buffer filled, we can assume there's some sort of repeat sequence and
; scan for the delay between the initial code and the repeat code, then find
; out how the repeat sequence is done and it's pattern.
; If the buffer didn't fill, we just save the length of the sequence & pattern.
; The area the data will go into is separated into 2 blocks - an area for the
; 16 bit pointers to each code, and the actual codes themselves. 
; The pointer area IRTABL first starts with the number of button codes, then 
; all the pointers to the offset from the beginning of the stored code area.
; The last (empty) pointer always points to the end of the IRCODE area where
; the next set of codes will be stored.
; The button codes are saved in the IRCODE area as follows:
;   first byte - length of the button code in bytes
;   second byte - length of the repeat code (00 if no repeat code)
;   remaining bytes - the actual button and repeat code timer reload values / 2
;-------------------------------------------------------------------------------

process_code:
	JNC	pc_cont		; if single button, fake a repeat time
	MOV	A,#080h		; with maximum count (about 140mS)
	MOVX	@DPTR,A
	INC	DPTR
	CLR	A
	MOVX	@DPTR,A
	DEC	R7
	
pc_cont:
	CLR	TR0		; stop the timer until it's re-configured
	POP 	TL0		; restore the previous setup of Timer 0
	POP	TH0		; as the time critical area is done
	POP	IE		; and let other routines have their INTs back
	POP	TMOD
	POP	TCON

	MOV	A,#0FFh		; find out how many bits were stored
	CLR	C
	SUBB	A,R7		; (2 * R7) + IR_BUF would point to buffer end
	MOV	R7,A		; save the length of the button bits if timeout
	MOV	R2,A		; save the actual length of the whole buffer
	MOV	R6,#0		; and the length of the repeat code
	CJNE	R7,#0,store_code	; if no repeat code, store what we got
					; else find out where the repeat is

;-------------------------------------------------------------------------------
; This next section scans the input buffer for delays between the button code
; and subsequent repeat codes.  It finds the maximum delay in the buffer, takes
; 87% of that for a good comparison (remotes usually use twice the longest bit)
; and starts the scan.  The first one is the button code delay, and it sets R7 
; to the end of the first delay.  The second one it finds is the delay between 
; repeat patterns, so the new value minus R7 is the length of the repeat code. 
; On nasty remotes like Yamaha that use 2 different delays, this will fail and
; either not find the second delay or miss the one after the button code.  You
; will need to hand-tweak the values in the buffer in that case.
; If there's any problem with the code missing delays, you can move the compare
; down to 75% by changing the divide factor just after comp_end.
; When we actually start looking for the code delay time, we need to skip the
; first few bits as some remotes send exceptionally long initial bits to wake
; up the receiver on the equipment.  I've seen 'em longer than the repeat delay
; in a couple of cases.  Once they're past, other bits are usually consistent.
;-------------------------------------------------------------------------------

find_repeat:			; have to scan data for significantly long delay
				; by finding (longest time - 13% ) or so
	MOV	DPTR,#(IR_BUF+8)	; move past the first bits (may be long as
					; most transmitters use cheap oscillators)
	MOV	R7,#0FAh		; set how many bits to check through
	MOVX	A,@DPTR		; start the scan - get a value
	MOV	max_delay,A	; save the current max msb
	INC	DPTR
	MOVX	A,@DPTR
	INC	DPTR
	MOV	max_delay+1,A	; save the current max lsb

comp_time:			; now start checking the buffer for the longest
	MOVX	A,@dptr		; delay between any sequence - usually pretty
	MOV	B,A		; consistent with keys from one remote, but
	INC	DPTR		; varies widely manufacturer to manufacturer
	MOVX	A,@DPTR		; as they all use different chips or schemes,
	INC	DPTR		; and not even the encoding is similar.
	DEC	R7
	CJNE	R7,#0,ct0	; if not at the last bit, keep going
	SJMP	comp_end

ct0:	XCH	A,B		; get the new msb in the acc & do 16-bit compare
	CJNE	A,max_delay,ct2	; check it against the old maximum value
	XCH	A,B		; if it fell thru, the MSBs are the same,
	CJNE	A,max_delay+1,ct1	; so check the LSBs
ct1:	XCH	A,B			; need to move msb back to B for later
ct2:	JC	comp_time	; if carry, new is less than previous - try again
	MOV	max_delay,A	; new is same or greater than previous
	MOV	max_delay+1,B	; so save it as the new max delay
	SJMP	comp_time	; and keep looking 'till we run out of bits

comp_end:			; now take 87% of the max delay for a new compare
	MOV	A,max_delay	; first get the old maximum into A,B
	MOV	B,max_delay+1
	MOV	R7,#3		; division factor - 3 for 13%, 2 for 25%
div1:	CLR	C		; and divide it by 8 (get 13% or thereabouts)
	RRC	A
	XCH	A,B		; division is handled with shifts because the
	RRC	A		; DIV AB instruction takes too much post-
	XCH	A,B		; processing to handle the remainder.  We don't
	DJNZ	R7,div1		; really need accuracy here, anyway.
	
	CLR	C		; now subtract the 13% from the real maximum
	XCH	A,max_delay	; and store the result back in max_delay
	XCH	A,B
	XCH	A,max_delay+1
	SUBB	A,max_delay+1
	MOV	max_delay+1,A
	MOV	A,B
	SUBB	A,max_delay	; now max_delay has 87% of it's old value
	MOV	max_delay,A	; so can be used to check for any repeat delays

	
	MOV	DPTR,#IR_BUF+8	; skip the first few bits - may be real long
	MOV	R0,#4		; adjust the word pointer past 'em
	ACALL	find_delay	; and find the length of the button sequence
	MOV	R7,A		; save for later storage
	RLC	A
	MOV	DPTR,#IR_BUF	; now move the DPTR to the end of button code
	ADD	A,DPL
	MOV	DPL,A
	CLR	A
	ADDC	A,DPH		; DPTR is now just past the button code delay
	MOV	DPH,A
	MOV	R0,#0		; didn't move past the start here, just init R0
	ACALL	find_delay	; and find the length of the repeat sequence
	MOV	R6,A		; save it likewise for storage later

;-------------------------------------------------------------------------------
;        All done - convert times, save the pointers and move to cold storage
;-------------------------------------------------------------------------------

store_code:
	MOV	A,R6		; first, let's convert the raw time codes to the
	ADD	A,R7		; actual packed timer reload values needed later
	MOV	DPTR,#IR_BUF
cvt_loop:
	PUSH	ACC		; and save for the moment
	MOVX	A,@DPTR		; get hi and low raw timer values
	MOV	B,A		; save hi byte in B
	INC	DPTR
	MOVX	A,@DPTR		; get low byte
	CLR	C
	RLC	A		; unpack into carry (was stored shifted)
	XCH	A,B		; swap hi and low bytes
	RLC	A		; finish unpacking into hi byte
	MOV	R0,A		; save hi byte in R0 for a sec
	MOV	F0,C		; save the overflow bit for a second
	CLR	C
	CLR	A		; convert to the needed reload values
	SUBB	A,B		; by subtracting the value from 1000h
	MOV	B,A
	CLR	A		; for both bytes
	SUBB	A,R0
	MOV	C,F0		; get the overflow bit back
	RRC	A		; pack back into high byte
	XCH	A,B
	RRC	A		; and repack low byte
	MOVX	@DPTR,A		; store back into the buffer - lo byte first
	DEC	DPL		;  since we didn't inc the DPTR yet
	MOV	A,#0FFh
	CJNE	A,DPL,convert_cont
	DEC	DPH
convert_cont:
	MOV	A,B
	MOVX	@DPTR,A		; then hi byte
	INC	DPTR
	INC	DPTR
	POP	ACC		; get the byte count back
	DJNZ	ACC,cvt_loop	; if more bytes in the table, convert 'em, too
;
; now store the converted data and update the pointer table
;
	ACALL	get_end_loc	; find the previous end of the buffer,
	MOVX	A,@DPTR		; now the start location for this code	
	PUSH	ACC		; got the high byte - stash it
	INC	DPTR
	MOVX	A,@DPTR		; got the low byte - put in DPL
	MOV	DPL,A
	POP	DPH		; DPTR now points to the first byte of storage

	MOV	A,R7		; get the length of the button code
	MOVX	@DPTR,A		; and write to storage
	INC	DPTR		;

	MOV	A,R6		; get length of the repeat code, even if none
	MOVX	@DPTR,A		; and write to storage
	INC	DPTR		;

	MOV	A,R7		; get the length of the button code
	MOV	P2,#high(IR_BUF)	; set the source location as P2,R0
	MOV	R0,#low(IR_BUF)		; and destination is still the DPTR
	ACALL	BLOCKMOV	; copy the button code to storage

	MOV	A,R6		; get the length of the repeat code
	JZ	savend		; if no repeat code, finish by saving end pointer
	ACALL	BLOCKMOV	; 'cos the block move will blow up with 0 length
	
savend:	PUSH	DPL		; save the last position
	PUSH	DPH
	MOV	DPTR,#IRTABL	; and move the button counter up a notch
	MOVX	A,@DPTR
	INC	A
	MOVX	@DPTR,A		; add back into the button counter

	ACALL	get_end_loc	; find the old end-of-data location
	POP	ACC		; get the last-used position back
	MOVX	@DPTR,A		; write the new end-of-data location 
	POP	ACC		; get the last-used lsb back
	INC	DPTR
	MOVX	@DPTR,A		; and write it, too	


alldone: 

	RET


;------------------------------------------------------------------------------
;                 SUBROUTINES USED BY THE MAINLINE CODE ABOVE
;------------------------------------------------------------------------------

get_end_loc:			; set DPTR at new entry area, ACC = # entries
	MOV	DPTR,#IRTABL	; find out how many entries are there already
	MOVX	A,@DPTR
	PUSH	ACC
	INC	DPTR
	JNZ	notfirst	; if it's not the first code ever, continue
	MOV	A,#high(IRCODE) ; else save the start of buffer at the pointer
	MOVX	@DPTR,A
	MOV	A,#low(IRCODE)
	INC	DPTR
	MOVX	@DPTR,A
	MOV	DPTR,#(IRTABL+1)
notfirst:
	POP	ACC		; get back the number of stored buttons
	PUSH	ACC		; save it for the end - just in case we need it
	JZ	fpe		; if it's the first one, stop here

fpl:	INC	DPTR		; move the DPTR to the beginning of the new code
	INC	DPTR
	DEC	A
	JNZ	fpl		; keep going till we hit the proper pointer

fpe:	POP	ACC		; and get the REAL button count back in ACC
	RET
	
;------------------------------------------------------------------------------

find_delay:		; SCAN INPUT BUFFER FOR INTER-CODE DELAY TIMES
	MOVX	A,@DPTR		; get the hi byte of the time
	MOV	B,A		; stash for a second
	INC	DPTR
	MOVX	A,@DPTR		; get the lo byte
	INC	DPTR
	INC	R0		; and up the byte counter
	XCH	A,B		; get the hi byte back
	CJNE	A,max_delay,fd2	; and see if it's greater than the delay
	XCH	A,B		; if it fell thru, the MSBs are the same,
	CJNE	A,max_delay+1,fd1	; so check the LSBs
fd1:	XCH	A,B
fd2:	JC	find_delay	; if carry, acc is less than max so retry
	CLR	C
	MOV	A,R0
	RET

;------------------------------------------------------------------------------
; move the code to RAM - or flash EPROM if you'd rather add some delays here

BLOCKMOV:			; MOVE SPECIFIED WORDS FROM ONE PLACE TO ANOTHER
	SETB	F0		; used so we move words, not bytes
bm:	PUSH	ACC		; P2/R0 have source addr, DPTR has dest addr
bm1:	MOVX	A,@R0		; and ACC has the number of bytes to move 
	MOVX	@DPTR,A		; if we're willing to trash a register or
	INC	DPTR		; data location somewhere, we can speed up the
	INC	R0		; block move by not PUSHing and POPping the ACC
	CJNE	R0,#0,bm2
	INC	P2		; however, it's not moving much data, so BFD

bm2:	JBC	F0,bm1
	SETB	F0
	POP	ACC

	DJNZ	ACC,bm		; all done moving the data yet?

	RET

end