/* CONSOLE.C -- serial I/O code                                              */

/*---------------------------------------------------------------------------*/
/* Initialize serial port hardware and variables                             */

SerInitialize() {

/*--- interrupt vectors                                                      */

 asm {
CONHERE EQU *
   ORG   $0023                   /* serial interrupt      (console I/O)      */
   LJMP  SerInt
   ORG   CONHERE
 }

   SerFlags = 0;                       /* flush bit flags to zero            */

   SETBIT(FlagTransEnable);
   SETBIT(FlagTransIdle);
   SETBIT(FlagAddCR);                  /* \n becomes CR/LF on output         */
   SETBIT(FlagStripLF);                /* discard incoming LFs               */

   CLRBIT(NET_ENABLE);                 /* disable transmitter                */

   RR_iHead = RR_iTail = RR_cLev = RR_cMax = 0;
   TR_iHead = TR_iTail = TR_cLev = TR_cMax = 0;

   UnGotCh = EOF;

/*--- set up Timer 1 to produce serial rate                                  */

   TCON &= 0x3F;                       /* clear run & interrupt flags        */

   TMOD &= 0x0F;                       /* flush existing Timer 1 setup       */
   TMOD |= 0x20;                       /* Timer 1 in 8-bit auto reload mode  */

   SCON = 0x50;                        /* serial mode 1, REN, TI off         */

   PCON |= 0x80;                       /* double the bit rate                */

   TH1 = TL1 = T1RELOAD & 0x00FF;      /* one-byte reload value              */

   SETBIT(TR1);                        /* start the timer                    */

   SETBIT(ES);                         /* enable serial interrupts           */


}


/*---------------------------------------------------------------------------*/
/* Serial console interrupt handler                                          */
/* If transmitter output is disabled, we fake trans interrupts until empty   */

SerInt() {


 asm {
*  SETB  SERIALFLAG

   PUSH  A                             ; save bystanders
   PUSH  PSW
   PUSH  DPH
   PUSH  DPL

*--- receiver interrupt

   JNB   RI,SItrans                    ; receiver interrupt active?

   MOV   A,RR_cLev                     ; room for newest char?
   CJAE  A,#INRINGSIZE,SIrend          ; no, so discard it

   INC   RR_cLev                       ; tick size counter
   MOV   A,RR_cLev                     ; track maximum value
   CJBE  A,RR_cMax,SIr1
   MOV   RR_cMax,A
SIr1  EQU   *

   MOV   DPTR,#RRing                   ; point to ring entry
   AddDPTR  RR_iHead

   MOV   A,SBUF                        ; pick up the character
   MOVX  [DPTR],A                      ; and stick in ring

   INC   RR_iHead                      ; tick the index
   MOV   A,RR_iHead                    ; hit end of array yet?
   CJB   A,#INRINGSIZE,SIrend
   MOV   RR_iHead,#0                   ; yes, reset to start

SIrend   EQU *
   CLR   RI                            ; indicate we have it


*--- transmitter interrupt

SItrans EQU *
   JNB   TI,SIdone                     ; transmitter interrupt active?

   MOV   A,TR_cLev                     ; anything to send?
   JNZ   SIt1
   SETB  FlagTransIdle                 ; no, flag inactive
   JNB   FlagUseTE,SItend              ; using network mode?
   CLR   A                             ; wait for a while
SItd  EQU   *
   DJNZ  A,SItd
   CLR   NET_ENABLE                    ; and turn off transmitter
   SJMP  SItend

SIt1  EQU *
   CLR   FlagTransIdle                 ; working on something
   JNB   FlagTransEnable,SItskip       ; output suppressed?
   JNB   FlagUseTE,SItex               ; using network mode?
   SETB  NET_ENABLE                    ; yes, so turn on transmitter
SItex EQU *
   MOV   DPTR,#TRing                   ; fetch next char
   AddDPTR   TR_iTail
   MOVX  A,[DPTR]
   MOV   SBUF,A                        ; and send it
SItskip  EQU *

   DEC   TR_cLev                       ; tick size counter

   INC   TR_iTail                      ; tick the index
   MOV   A,TR_iTail                    ; hit end of array yet?
   CJB   A,#OUTRINGSIZE,SItend
   MOV   TR_iTail,#0                   ; yes, reset to start

SItend    EQU   *
   CLR   TI                            ; indicate done with it

   JB    FlagTransEnable,SIdone        ; output suppressed?
   JB    FlagTransIdle,SIdone          ; transmitter finished?
   SETB  TI                            ; no to both -> fake transmission

*--- clean up & go home

SIdone   EQU *

   POP   DPL                           ; restore bystanders
   POP   DPH
   POP   PSW
   POP   A

*  CLR   SERIALFLAG

   RETI                                ; preempt normal return

 }


}

/*---------------------------------------------------------------------------*/
/* Send character to console                                                 */
/* May add CR in front of each LF to provide standard DOS sequence           */

putch(TransChar)
int TransChar;
{

 asm {

   MOV   R0,#-5                        ; get character from stack
   LCALL ?auto0
   MOV   A,[R0]

?putch EQU *                           ; vital for library routines!
   PUSH  A                             ; save char for later

   CJNE  A,#LF,pch1                    ; if not LF, handle normally
   JNB   FlagAddCR,pch1                ; if not adding CRs, handle normally
   SETB  FlagLineEndOut                ; remember we have a line end active
   MOV   A,#CR                         ; substitute a CR for the LF
   DEC   SP                            ; replace PUSHed char
   PUSH  A                             ;  ... with the new CR

pch1  EQU   *
   MOV   A,TR_cLev                     ; wait for space in ring
*  CPL   SERIALFLAG
*  CPL   SERIALFLAG
   CJAE  A,#OUTRINGSIZE,pch1

   CLR   ES

   MOV   DPTR,#TRing                   ; point to char slot
   AddDPTR TR_iHead
   POP   A                             ; recover character
   PUSH  A                             ; save again
   MOVX  [DPTR],A

   INC   TR_cLev                       ; tick counter & index
   INC   TR_iHead

   MOV   A,TR_cLev                     ; update maximum level
   CJB   A,TR_cMax,pch2
   MOV   TR_cMax,A
pch2  EQU   *

   MOV   A,TR_iHead
   CJB   A,#OUTRINGSIZE,pch3
   MOV   TR_iHead,#0
pch3  EQU   *

   JNB   FlagTransIdle,pch4            ; kickstart transmitter if idle
   SETB  TI

pch4

   SETB  ES

   POP   A                             ; recover original char

   JNB   FlagLineEndOut,pch5           ; did we just send the CR of a CR/LF?
   CLR   FlagLineEndOut                ; clear flag to indicate it is done
   MOV   A,#LF                         ; yes, so we must add the LF
   PUSH  A                             ; make stack come out right
   SJMP  pch1                          ; and send this one off

pch5  EQU   *
   RET                                 ; bypass dead code

 }

   TransChar = 0;                /* dummy to refer to argument               */

}


/*---------------------------------------------------------------------------*/
/* Decide if there are any pending chars                                     */
/* Returns nonzero if there's a char                                         */
/* Discards LFs if that flag is set                                          */
/* Translates CRs into LFs (\n chars) if we are discarding LFs               */

int chkch() {

 asm {

   MOV   A,UnGotCh                     ; anything saved?
   XRL   A,#EOF                        ; nonzero if different from EOF
   JNZ   CC9                           ; if present, return without processing

CC1   EQU   *
   MOV   A,RR_cLev                     ; anything in the incoming ring?
   JZ    CC9                           ; if no char, we are done here
   JNB   FlagStripLF,CC9               ; if not stripping LFs, avoid tests

   CLR   ES                            ; peek at the character
   MOV   DPTR,#RRing
   AddDPTR  RR_iTail
   MOVX  A,[DPTR]
   SETB  ES

   CJE   A,#LF,CC2                     ; if it is a LF, discard it
   MOV   A,#1                          ; otherwise, return a nonzero!
   SJMP  CC9

CC2   EQU   *
   DEC   RR_cLev                       ; pull the LF out of the ring

   INC   RR_iTail                      ; tick the index
   MOV   A,RR_iTail                    ; hit end of array yet?
   CJB   A,#INRINGSIZE,CC4
   MOV   RR_iTail,#0                   ; yes, reset to start

CC4   EQU   *
   SJMP  CC1                           ; see if next char is any better!

CC9   EQU   *
   MOV   B,#0                          ; clear high byte

 }

}

/*---------------------------------------------------------------------------*/
/* Wait for the serial transmitter to go idle                                */
/* If the transmitter is disabled, that's considered to be the same thing    */

SerWaitOutDone() {

   TESTBIT(TestBits,FlagTransEnable);
   if (TestBits) {
      while (TR_cLev) {                            /* wait for ring empty    */
         continue;
      }

      do {
         TESTBIT(TestBits,FlagTransIdle);
      } while (!TestBits);                         /* wait for last char     */
   }

}

/*---------------------------------------------------------------------------*/
/* Flush the input buffer                                                    */
/* Returns number of chars flushed                                           */

int SerFlushIn() {

 asm {

   CLR   ES                            ; turn off serial interrupts

   MOV   UnGotCh,#EOF                  ; nothing saved

   MOV   A,#0                          ; pick up a zero
   MOV   RR_iTail,A                    ; reset ring variables
   MOV   RR_iHead,A                    ;  ... clapping head & tail together
   XCH   A,RR_cLev                     ; return # chars flushed, leave a zero

   SETB  ES                            ; enable serial again

   MOV   B,#0                          ; ensure MSB is zero

 }

}


/*---------------------------------------------------------------------------*/
/* Get character from console                                                */
/* FlagStripLF forces LFs to disappear on input                              */

int getch() {

 asm {

*--- have a saved char?

   MOV   A,#EOF                        ; scratch saved char regardless
   XCH   A,UnGotCh
   CJNE  A,#EOF,GCpolish               ; had one, so fix it up for use

*--- check for serial port char

GCwait   EQU *

   MOV   A,RR_cLev                     ; anything pending?
*  CPL   SERIALFLAG
   NOP
*  CPL   SERIALFLAG
   JZ    GCwait                        ; not when count is zero

   CLR   ES                            ; avoid interruptions

   MOV   DPTR,#RRing                   ; fetch char
   AddDPTR  RR_iTail
   MOVX  A,[DPTR]

   PUSH  A                             ; save it for later

   DEC   RR_cLev                       ; tick size counter

   INC   RR_iTail                      ; tick the index
   MOV   A,RR_iTail                    ; hit end of array yet?
   CJB   A,#INRINGSIZE,GC2
   MOV   RR_iTail,#0                   ; yes, reset to start
GC2 EQU *

   SETB  ES                            ; interruptions OK now

   POP   A                             ; recover char

*--- decide whether to return this char or convert it to nothing at all
*    chchk uses this same logic to convert & discard pending chars, too
*     ... so we never discard a char that was reported by chkch as present

GCpolish EQU *

   JNB   FlagStripLF,GC9               : are we stripping LFs?
   CJE   A,#LF,GCWait                  ; if LF, go back for next char

   CJNE  A,#CR,GC9                     ; if not a CR, pass through
   MOV   A,#LF                         ;  ... else, convert to \n for C code

GC9   EQU   *
   MOV   B,#0

 }
}



/*---------------------------------------------------------------------------*/
/* Get character from console with echo                                      */

int getche() {

 asm {
   LCALL getch                         ; fetch character w/o echo
   PUSH  A                             ; save it for a while

   PUSH  A                             ; push 2-byte char for sure!
   PUSH  A
   LCALL putch                         ; and display it
   POP   A                             ; discard stacked parameter
   POP   A

   POP   A                             ; recover char
   MOV   B,#0                          ; set up for return

 }
}


/*---------------------------------------------------------------------------*/
/* Push character back into console                                          */
/* This returns the char or EOF if there was one undone already              */

ungetch(XChar)
int XChar;
{

   if (EOF == UnGotCh) {
      return EOF;
   }
   else {
      UnGotCh = XChar;
      return XChar;
   }

}


/*---------------------------------------------------------------------------*/
/* Send string to console                                                    */
/*  putstr(char *pString);                                                   */
/*  The ?putstr entry point has *pString in DPTR                             */

putstr(pString)
char *pString;
{
 asm {
   MOV   R0,#-5            ; get string pointer from stack
   LCALL ?auto0
   MOV   DPL,[R0]          ; low byte
   INC   R0
   MOV   DPH,[R0]
?putstr
   MOVX  A,[DPTR]          ; fetch character
   JZ    pstr1             ; if zero, done
   PUSH  DPH               ; save pointer
   PUSH  DPL
   PUSH  A                 ; save argument
   PUSH  A
   LCALL putch             ; send it out
   POP   A                 ; discard parameter
   POP   A
   POP   DPL               ; restore pointer
   POP   DPH
   INC   DPTR
   SJMP  ?putstr           ; continue...

pstr1 EQU *
   RET

 }

   pString = 0;            /* dummy code to reference pString...             */

}



/*---------------------------------------------------------------------------*/
/* Dump serial port status                                                   */
/* Side effect resets maximum ring counters                                  */

SerDumpState() {

   printf("Console status flags %02x\n",SerFlags);

   printf(" Trans:  Head %02x, Tail %02x, Ctr %02x, Max %02x\n",
          TR_iHead,TR_iTail,TR_cLev,TR_cMax);
   printf(" Rec:    Head %02x, Tail %02x, Ctr %02x, Max %02x\n",
          RR_iHead,RR_iTail,RR_cLev,RR_cMax);

   TR_cMax = RR_cMax = 0;

}


/*---------------------------------------------------------------------------*/
/* Convert string to hex                                                     */
/* Note that ppHexits is a pointer to the pointer to the string...           */
/*  so we can return the pointer incremented to just beyond the last char    */
/*  we use in the conversion...                                              */
/* Must be in external RAM for Compact model, thus be careful about local    */
/*  variables and so forth...                                                */

#define SHOWGETHEXxxx

WORD GetHexits(ppHexits,MaxLen)
char **ppHexits;
unsigned MaxLen;
{
WORD Result;
char HexChar;

   Result = 0;


#ifdef SHOWGETHEX
   printf("Entering GetHexits, arg %04x, *arg %04x **arg %04x\n",
      ppHexits,*ppHexits,**ppHexits);
#endif

   while ((MaxLen--) && isxdigit(**ppHexits)) {
      Result <<= 4;
      HexChar = toupper(**ppHexits);
#ifdef SHOWGETHEX
      printf("**ppHexits is [%c] and HexChar is [%c]\n",**ppHexits,HexChar);
#endif
      Result += (HexChar <= '9') ? (HexChar - '0') : (HexChar - 'A' + 10);
      ++(*ppHexits);
   }

#ifdef SHOWGETHEX
   printf("Result is %04x\n",Result);
#endif

   return Result;

}



/*---------------------------------------------------------------------------*/
/* Convert string to decimal number                                          */
/* Note that ppDigits is a pointer to the pointer to the string...           */
/*  so we can return the pointer incremented to just beyond the last char    */
/*  we use in the conversion...                                              */

WORD GetDecimal(ppDigits)
char **ppDigits;
{
WORD Result;
char Buffer[10];
char *pBuffer;
BYTE Len;

   memset(Buffer,0,sizeof(Buffer));
   pBuffer = Buffer;
   Len = sizeof(Buffer) - 1;

   while (isascii(**ppDigits) && isdigit(**ppDigits) && Len) {
      *(pBuffer++) = **ppDigits;
      ++(*ppDigits);
      --Len;
   }

   Result = atoi(Buffer);

}
